import locale
import datetime

from flask import abort
from flask import Blueprint
from flask import current_app
from flask import jsonify
from flask import render_template
from flask import request

from . import cache

from .search import get_date, get_sort

bp = Blueprint("administration", __name__)


@bp.route('/administration/<administration_id>', methods=['GET'])
@cache.cached(timeout=0, query_string=True)
def list_raa(administration_id):
    start_date = get_date(request.args.get('start_date', '', type=str))
    end_date = get_date(request.args.get('end_date', '', type=str))
    sort = get_sort(request.args.get('sort', '', type=str))
    if sort == 'relevance':
        sort = 'desc'

    size = 50
    try:
        size = int(request.args.get('size', '50'))
        if size > 300:
            size = 300
    except:
        size = 50

    page = 1
    try:
        page = int(request.args.get('page', '1'))
        if page <= 1:
            page = 1
    except:
        page = 1

    raa = list_raa_from_opensearch(administration_id, size=size, page=page, start_date=start_date, end_date=end_date, sort=sort)
    if len(raa['elements']) == 0:
        return abort(404)

    try:
        locale.setlocale(locale.LC_ALL, 'fr_FR.UTF-8')
    except Exception:
        pass

    first_raa_date = datetime.datetime.strptime(list_raa_from_opensearch(administration_id, size=1, sort='asc')['elements'][0]['date'], '%Y-%m-%d').strftime('%d %B %Y')
    last_raa_date = datetime.datetime.strptime(list_raa_from_opensearch(administration_id, size=1, sort='desc')['elements'][0]['date'], '%Y-%m-%d').strftime('%d %B %Y')

    # On transforme la date en quelque chose de lisible
    elements = []
    for element in raa['elements']:
        element['date'] = datetime.datetime.strptime(element['date'], '%Y-%m-%d').strftime('%d %B %Y')
        elements.append(element)
    raa['elements'] = elements

    return render_template(
        'administration.html',
        raa=raa,
        administration_id=administration_id,
        administration_name=current_app.config['AVAILABLE_ADMINISTRATIONS'][administration_id],
        size=size,
        page=page,
        start_date=start_date,
        end_date=end_date,
        sort=sort,
        first_raa_date=first_raa_date,
        last_raa_date=last_raa_date
    )

@bp.route('/api/v1/administration/<administration_id>', methods=['GET'])
@cache.cached(timeout=0, query_string=True)
def api_v1_list_raa(administration_id):
    start_date = get_date(request.args.get('start_date', '', type=str))
    end_date = get_date(request.args.get('end_date', '', type=str))
    sort = get_sort(request.args.get('sort', '', type=str))
    if sort == 'relevance':
        sort = 'desc'

    size = 50
    try:
        size = int(request.args.get('size', '50'))
        if size > 300:
            size = 300
    except:
        size = 50

    page = 1
    try:
        page = int(request.args.get('page', '1'))
        if page <= 1:
            page = 1
    except:
        page = 1

    raa = list_raa_from_opensearch(administration_id, size=size, page=page, start_date=start_date, end_date=end_date, sort=sort)
    if len(raa['elements']) == 0:
        return abort(404)
    else:
        return jsonify(raa)

def list_raa_from_opensearch(administration_id, size=10, page=1, start_date=None, end_date=None, sort='desc'):
    if not administration_id or administration_id == '' or not administration_id in current_app.config['AVAILABLE_ADMINISTRATIONS']:
        result = {
            'total': 0,
            'elements': []
        }
        return result
    else:
        query = {
            'from': (page-1)*size,
            'size': size,
            'query': {
                'bool': {
                    'must': {
                        'term': {
                            'administration': administration_id
                        }
                    }
                }
            }
        }

        if start_date is not None:
            start_date_filter = {
                'range': {
                    'date': {}
                }
            }
            start_date_filter['range']['date']['gte'] = start_date
            query['query']['bool']['must'].append(start_date_filter)

        if end_date is not None:
            end_date_filter = {
                'range': {
                    'date': {}
                }
            }
            end_date_filter['range']['date']['lte'] = end_date
            query['query']['bool']['must'].append(end_date_filter)

        if sort == 'asc' or sort == 'desc':
            query['sort'] = [
                {
                    'date': {
                        'order': sort,
                    }
                },
                {
                    'first_seen_on': {
                        'order': sort,
                    }
                }
            ]
        if sort == 'first_seen_on_desc' or sort == 'first_seen_on_asc':
            if sort == 'first_seen_on_desc':
                order = 'desc'
            if sort == 'first_seen_on_asc':
                order = 'asc'
            query['sort'] = [
                {
                    'first_seen_on': {
                        'order': order,
                    }
                }
            ]

        response = current_app.opensearch.search(
            body = query,
            index = 'attrap-index'
        )

        result = {
            'total': int(response['hits']['total']['value']),
            'elements': []
        }

        for raa in response['hits']['hits']:
            del(raa['_source']['content'])
            del(raa['_source']['administration'])
            del(raa['_source']['url'])
            del(raa['_source']['timezone'])
            del(raa['_source']['pdf_creation_date'])
            del(raa['_source']['pdf_modification_date'])

            result['elements'].append(raa['_source'] | { 'id': raa['_id'] })

        return result
